"""
Qnty - High-Performance Unit System for Engineering
====================================================

A fast, type-safe unit system library for Python with dimensional safety
and optimized unit conversions for engineering calculations.
"""

from .dimension import BaseDimension, DimensionSignature
from .equation import Equation
from .expression import Expression
from .unit import registry
from .units import register_all_units
from .variable import FastQuantity, TypeSafeSetter, TypeSafeVariable
from .variables import (
    AbsorbedDose,
    Acceleration,
    ActivationEnergy,
    AmountOfSubstance,
    AnglePlane,
    AngleSolid,
    AngularAcceleration,
    AngularMomentum,
    Area,
    AreaPerUnitVolume,
    AtomicWeight,
    Concentration,
    DynamicFluidity,
    ElectricCapacitance,
    ElectricCharge,
    ElectricCurrentIntensity,
    ElectricDipoleMoment,
    ElectricFieldStrength,
    ElectricInductance,
    ElectricPotential,
    ElectricResistance,
    ElectricalConductance,
    ElectricalPermittivity,
    ElectricalResistivity,
    EnergyFlux,
    EnergyHeatWork,
    EnergyPerUnitArea,
    Force,
    ForceBody,
    ForcePerUnitMass,
    FrequencyVoltageRatio,
    FuelConsumption,
    HeatOfCombustion,
    HeatOfFusion,
    HeatOfVaporization,
    HeatTransferCoefficient,
    Illuminance,
    KineticEnergyOfTurbulence,
    Length,
    LinearMassDensity,
    LinearMomentum,
    LuminanceSelf,
    LuminousFlux,
    LuminousIntensity,
    MagneticField,
    MagneticFlux,
    MagneticInductionFieldStrength,
    MagneticMoment,
    MagneticPermeability,
    MagnetomotiveForce,
    Mass,
    MassDensity,
    MassFlowRate,
    MassFlux,
    MassFractionOfI,
    MassTransferCoefficient,
    MolalityOfSoluteI,
    MolarConcentrationByMass,
    MolarFlowRate,
    MolarFlux,
    MolarHeatCapacity,
    MolarityOfI,
    MoleFractionOfI,
    MomentOfInertia,
    MomentumFlowRate,
    MomentumFlux,
    NormalityOfSolution,
    ParticleDensity,
    Permeability,
    PhotonEmissionRate,
    PowerPerUnitMass,
    PowerPerUnitVolume,
    PowerThermalDuty,
    Pressure,
    RadiationDoseEquivalent,
    RadiationExposure,
    Radioactivity,
    SecondMomentOfArea,
    SecondRadiationConstantPlanck,
    SpecificEnthalpy,
    SpecificGravity,
    SpecificHeatCapacityConstantPressure,
    SpecificLength,
    SpecificSurface,
    SpecificVolume,
    Stress,
    SurfaceMassDensity,
    SurfaceTension,
    Temperature,
    ThermalConductivity,
    Time,
    Torque,
    TurbulenceEnergyDissipationRate,
    VelocityAngular,
    VelocityLinear,
    ViscosityDynamic,
    ViscosityKinematic,
    Volume,
    VolumeFractionOfI,
    VolumetricCalorificHeatingValue,
    VolumetricCoefficientOfExpansion,
    VolumetricFlowRate,
    VolumetricFlux,
    VolumetricMassFlowRate,
    Wavenumber
)

# Register all units to the global registry
register_all_units(registry)

# Finalize registry after all registrations
registry.finalize_registration()

# Version information
__version__ = "0.0.3"

# Define public API
__all__ = [
    # Core variable types (most commonly used)
    "Length", "Pressure", "Temperature", "Time", "Mass", "Volume", "Area",
    "Force", "EnergyHeatWork", "PowerThermalDuty",

    # Core classes for advanced usage
    "FastQuantity", "TypeSafeVariable", "TypeSafeSetter",
    "DimensionSignature", "BaseDimension",
    "Expression", "Equation",

    # All other variable types (95 additional types)
    "AbsorbedDose", "Acceleration", "ActivationEnergy", "AmountOfSubstance",
    "AnglePlane", "AngleSolid", "AngularAcceleration", "AngularMomentum",
    "AreaPerUnitVolume", "AtomicWeight", "Concentration", "DynamicFluidity",
    "ElectricCapacitance", "ElectricCharge", "ElectricCurrentIntensity",
    "ElectricDipoleMoment", "ElectricFieldStrength", "ElectricInductance",
    "ElectricPotential", "ElectricResistance", "ElectricalConductance",
    "ElectricalPermittivity", "ElectricalResistivity", "EnergyFlux",
    "EnergyPerUnitArea", "ForceBody", "ForcePerUnitMass",
    "FrequencyVoltageRatio", "FuelConsumption", "HeatOfCombustion",
    "HeatOfFusion", "HeatOfVaporization", "HeatTransferCoefficient",
    "Illuminance", "KineticEnergyOfTurbulence", "LinearMassDensity",
    "LinearMomentum", "LuminanceSelf", "LuminousFlux", "LuminousIntensity",
    "MagneticField", "MagneticFlux", "MagneticInductionFieldStrength",
    "MagneticMoment", "MagneticPermeability", "MagnetomotiveForce",
    "MassDensity", "MassFlowRate", "MassFlux", "MassFractionOfI",
    "MassTransferCoefficient", "MolalityOfSoluteI", "MolarConcentrationByMass",
    "MolarFlowRate", "MolarFlux", "MolarHeatCapacity", "MolarityOfI",
    "MoleFractionOfI", "MomentOfInertia", "MomentumFlowRate", "MomentumFlux",
    "NormalityOfSolution", "ParticleDensity", "Permeability",
    "PhotonEmissionRate", "PowerPerUnitMass", "PowerPerUnitVolume",
    "RadiationDoseEquivalent", "RadiationExposure", "Radioactivity",
    "SecondMomentOfArea", "SecondRadiationConstantPlanck", "SpecificEnthalpy",
    "SpecificGravity", "SpecificHeatCapacityConstantPressure",
    "SpecificLength", "SpecificSurface", "SpecificVolume", "Stress",
    "SurfaceMassDensity", "SurfaceTension", "ThermalConductivity", "Torque",
    "TurbulenceEnergyDissipationRate", "VelocityAngular", "VelocityLinear",
    "ViscosityDynamic", "ViscosityKinematic", "VolumeFractionOfI",
    "VolumetricCalorificHeatingValue", "VolumetricCoefficientOfExpansion",
    "VolumetricFlowRate", "VolumetricFlux", "VolumetricMassFlowRate",
    "Wavenumber",
]




"""
SI Prefix System
================

Standard SI prefixes for unit multiplication/division.
Provides systematic handling of metric prefixes like kilo-, milli-, micro-, etc.
"""

from dataclasses import dataclass
from enum import Enum


@dataclass(frozen=True)
class SIPrefix:
    """
    Standard SI prefix definition.
    
    Attributes:
        name: Full prefix name (e.g., "kilo", "milli")
        symbol: Standard symbol (e.g., "k", "m")
        factor: Multiplication factor relative to base unit (e.g., 1000, 0.001)
    """
    name: str
    symbol: str
    factor: float
    
    def apply_to_name(self, base_name: str) -> str:
        """Apply prefix to a base unit name."""
        if not self.name:
            return base_name
        return f"{self.name}{base_name}"
    
    def apply_to_symbol(self, base_symbol: str) -> str:
        """Apply prefix to a base unit symbol."""
        if not self.symbol:
            return base_symbol
        return f"{self.symbol}{base_symbol}"


class StandardPrefixes(Enum):
    """
    Standard SI prefixes with their multiplication factors.
    
    From yotta (10^24) to yocto (10^-24).
    """
    # Larger prefixes (10^3 to 10^24)
    YOTTA = SIPrefix("yotta", "Y", 1e24)
    ZETTA = SIPrefix("zetta", "Z", 1e21)
    EXA = SIPrefix("exa", "E", 1e18)
    PETA = SIPrefix("peta", "P", 1e15)
    TERA = SIPrefix("tera", "T", 1e12)
    GIGA = SIPrefix("giga", "G", 1e9)
    MEGA = SIPrefix("mega", "M", 1e6)
    KILO = SIPrefix("kilo", "k", 1e3)
    HECTO = SIPrefix("hecto", "h", 1e2)
    DECA = SIPrefix("deca", "da", 1e1)
    
    # Base (no prefix)
    NONE = SIPrefix("", "", 1.0)
    
    # Smaller prefixes (10^-1 to 10^-24)
    DECI = SIPrefix("deci", "d", 1e-1)
    CENTI = SIPrefix("centi", "c", 1e-2)
    MILLI = SIPrefix("milli", "m", 1e-3)
    MICRO = SIPrefix("micro", "μ", 1e-6)
    NANO = SIPrefix("nano", "n", 1e-9)
    PICO = SIPrefix("pico", "p", 1e-12)
    FEMTO = SIPrefix("femto", "f", 1e-15)
    ATTO = SIPrefix("atto", "a", 1e-18)
    ZEPTO = SIPrefix("zepto", "z", 1e-21)
    YOCTO = SIPrefix("yocto", "y", 1e-24)


# Common prefix groups for different unit types
COMMON_LENGTH_PREFIXES = [
    StandardPrefixes.KILO,
    StandardPrefixes.CENTI,
    StandardPrefixes.MILLI,
    StandardPrefixes.MICRO,
    StandardPrefixes.NANO,
]

COMMON_MASS_PREFIXES = [
    StandardPrefixes.KILO,  # Note: kilogram is the SI base unit
    StandardPrefixes.MILLI,
    StandardPrefixes.MICRO,
]

COMMON_TIME_PREFIXES = [
    StandardPrefixes.MILLI,
    StandardPrefixes.MICRO,
    StandardPrefixes.NANO,
    StandardPrefixes.PICO,
]

COMMON_ELECTRIC_PREFIXES = [
    StandardPrefixes.KILO,
    StandardPrefixes.MILLI,
    StandardPrefixes.MICRO,
    StandardPrefixes.NANO,
    StandardPrefixes.PICO,
]

COMMON_ENERGY_PREFIXES = [
    StandardPrefixes.KILO,
    StandardPrefixes.MEGA,
    StandardPrefixes.GIGA,
]

COMMON_POWER_PREFIXES = [
    StandardPrefixes.KILO,
    StandardPrefixes.MEGA,
    StandardPrefixes.GIGA,
    StandardPrefixes.MILLI,
    StandardPrefixes.MICRO,
]

COMMON_PRESSURE_PREFIXES = [
    StandardPrefixes.KILO,
    StandardPrefixes.MEGA,
    StandardPrefixes.GIGA,
]


def get_prefix_by_name(name: str) -> SIPrefix | None:
    """Get a prefix by its name (e.g., 'kilo', 'milli')."""
    for prefix_enum in StandardPrefixes:
        if prefix_enum.value.name == name:
            return prefix_enum.value
    return None


def get_prefix_by_symbol(symbol: str) -> SIPrefix | None:
    """Get a prefix by its symbol (e.g., 'k', 'm')."""
    for prefix_enum in StandardPrefixes:
        if prefix_enum.value.symbol == symbol:
            return prefix_enum.value
    return None


def get_prefix_by_factor(factor: float, tolerance: float = 1e-10) -> SIPrefix | None:
    """Get a prefix by its multiplication factor."""
    for prefix_enum in StandardPrefixes:
        if abs(prefix_enum.value.factor - factor) < tolerance:
            return prefix_enum.value
    return None


# Define which units should get automatic prefixes
PREFIXABLE_UNITS = {
    # Base SI units
    'meter': COMMON_LENGTH_PREFIXES,
    'gram': COMMON_MASS_PREFIXES,
    'second': COMMON_TIME_PREFIXES,
    'ampere': COMMON_ELECTRIC_PREFIXES,
    'kelvin': [],  # Temperature usually doesn't use prefixes
    'mole': [
        StandardPrefixes.MILLI,
        StandardPrefixes.MICRO
    ],
    'candela': [],  # Luminous intensity rarely uses prefixes
    
    # Derived SI units
    'pascal': COMMON_PRESSURE_PREFIXES,
    'joule': COMMON_ENERGY_PREFIXES,
    'watt': COMMON_POWER_PREFIXES,
    'coulomb': COMMON_ELECTRIC_PREFIXES,
    'volt': COMMON_ELECTRIC_PREFIXES,
    'farad': [
        StandardPrefixes.MILLI,
        StandardPrefixes.MICRO,
        StandardPrefixes.NANO,
        StandardPrefixes.PICO
    ],
    'ohm': [
        StandardPrefixes.KILO,
        StandardPrefixes.MEGA,
        StandardPrefixes.MILLI
    ],
    'siemens': [
        StandardPrefixes.MILLI,
        StandardPrefixes.MICRO
    ],
    'weber': [
        StandardPrefixes.MILLI,
        StandardPrefixes.MICRO
    ],
    'tesla': [
        StandardPrefixes.MILLI,
        StandardPrefixes.MICRO,
        StandardPrefixes.NANO
    ],
    'henry': [
        StandardPrefixes.MILLI,
        StandardPrefixes.MICRO,
        StandardPrefixes.NANO
    ],
    'lumen': [],
    'lux': [],
    'becquerel': [
        StandardPrefixes.KILO,
        StandardPrefixes.MEGA,
        StandardPrefixes.GIGA
    ],
    'gray': [
        StandardPrefixes.MILLI,
        StandardPrefixes.MICRO
    ],
    'sievert': [
        StandardPrefixes.MILLI,
        StandardPrefixes.MICRO
    ],
    'hertz': [
        StandardPrefixes.KILO,
        StandardPrefixes.MEGA,
        StandardPrefixes.GIGA
    ],
    'newton': [
        StandardPrefixes.KILO,
        StandardPrefixes.MILLI
    ],
    'bar': [
        StandardPrefixes.MILLI
    ],  # Common non-SI unit
    'liter': [
        StandardPrefixes.MILLI,
        StandardPrefixes.MICRO
    ],  # Common non-SI unit
}



"""
Expression Variable Base Class
==============================

Base class that extends TypeSafeVariable with mathematical expression
and equation capabilities.
"""

from __future__ import annotations

from ..equation import Equation
from ..expression import Expression, wrap_operand
from ..variable import FastQuantity, TypeSafeVariable


class ExpressionVariable(TypeSafeVariable):
    """
    TypeSafeVariable extended with expression and equation capabilities.
    
    This adds mathematical operations that create expressions and equations,
    keeping the base TypeSafeVariable free of these dependencies.
    """
    
    def equals(self, expression: Expression | TypeSafeVariable | FastQuantity | int | float):
        """Create an equation: self = expression."""
        # Wrap the expression in proper Expression type
        rhs_expr = wrap_operand(expression)
        return Equation(f"{self.name}_eq", self, rhs_expr)
    
    def __add__(self, other: TypeSafeVariable | FastQuantity | int | float) -> Expression:
        """Add this variable to another operand, returning an Expression."""
        return wrap_operand(self) + wrap_operand(other)

    def __radd__(self, other: FastQuantity | int | float) -> Expression:
        """Reverse add for this variable."""
        return wrap_operand(other) + wrap_operand(self)
    
    def __sub__(self, other: TypeSafeVariable | FastQuantity | int | float) -> Expression:
        """Subtract another operand from this variable, returning an Expression."""
        return wrap_operand(self) - wrap_operand(other)
    
    def __rsub__(self, other: FastQuantity | int | float) -> Expression:
        """Reverse subtract for this variable."""
        return wrap_operand(other) - wrap_operand(self)
    
    def __mul__(self, other: TypeSafeVariable | FastQuantity | int | float) -> Expression:
        """Multiply this variable by another operand, returning an Expression."""
        return wrap_operand(self) * wrap_operand(other)
    
    def __rmul__(self, other: FastQuantity | int | float) -> Expression:
        """Reverse multiply for this variable."""
        return wrap_operand(other) * wrap_operand(self)
    
    def __truediv__(self, other: TypeSafeVariable | FastQuantity | int | float) -> Expression:
        """Divide this variable by another operand, returning an Expression."""
        return wrap_operand(self) / wrap_operand(other)

    def __rtruediv__(self, other: FastQuantity | int | float) -> Expression:
        """Reverse divide for this variable."""
        return wrap_operand(other) / wrap_operand(self)
    
    def __pow__(self, other: TypeSafeVariable | FastQuantity | int | float) -> Expression:
        """Raise this variable to a power, returning an Expression."""
        return wrap_operand(self) ** wrap_operand(other)

    def __rpow__(self, other: FastQuantity | int | float) -> Expression:
        """Reverse power for this variable."""
        return wrap_operand(other) ** wrap_operand(self)



"""
Base Unit Module Definition
===========================

Provides abstract base class for unit modules and registration functionality.
"""

from abc import ABC, abstractmethod
from typing import Any

from ..unit import UnitDefinition


class UnitModule(ABC):
    """Abstract base class for unit modules."""
    
    @abstractmethod
    def get_unit_definitions(self) -> list[UnitDefinition]:
        """Return list of unit definitions for this module."""
        pass
    
    @abstractmethod
    def get_units_class(self) -> type[Any]:
        """Return the units class for this module."""
        pass
    
    def register_to_registry(self, unit_registry):
        """Register all unit definitions to the given registry."""
        from ..unit import UnitConstant
        
        for unit_def in self.get_unit_definitions():
            if unit_def.name not in unit_registry.units:
                unit_registry.register_unit(unit_def)
        
        # Populate the units class with constants from registry
        units_class = self.get_units_class()
        for unit_def in self.get_unit_definitions():
            unit_constant = UnitConstant(unit_registry.units[unit_def.name])
            setattr(units_class, unit_def.name, unit_constant)
            
            # Add any aliases
            if unit_def.symbol and unit_def.symbol != unit_def.name:
                setattr(units_class, unit_def.symbol, unit_constant)
            
            # Special case for inch - add in_ alias since 'in' is a Python keyword
            if unit_def.name == "inch":
                units_class.in_ = unit_constant
